@@ -31,4 +31,13 @@ public interface Vector extends Serializable {
 
     /** Converts the instance to a double array. */
     double[] toArray();
+
+    /** Converts the instance to a dense vector. */
+    DenseVector toDense();
+
+    /** Converts the instance to a sparse vector. */
+    SparseVector toSparse();
+
+    /** Makes a deep copy of the vector. */
+    Vector clone();
 }
