@@ -24,31 +24,36 @@ import org.apache.flink.api.connector.source.Source;
 import org.apache.flink.connector.file.sink.FileSink;
 import org.apache.flink.connector.file.src.FileSource;
 import org.apache.flink.connector.file.src.reader.SimpleStreamFormat;
+import org.apache.flink.core.fs.FileSystem;
+import org.apache.flink.core.fs.Path;
 import org.apache.flink.ml.api.Stage;
 import org.apache.flink.ml.builder.Graph;
 import org.apache.flink.ml.builder.GraphData;
 import org.apache.flink.ml.builder.GraphModel;
 import org.apache.flink.ml.builder.GraphNode;
+import org.apache.flink.ml.common.datastream.TableUtils;
 import org.apache.flink.ml.param.Param;
+import org.apache.flink.ml.param.WithParams;
 import org.apache.flink.streaming.api.datastream.DataStream;
 import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
 import org.apache.flink.streaming.api.functions.sink.filesystem.bucketassigners.BasePathBucketAssigner;
 import org.apache.flink.streaming.api.functions.sink.filesystem.rollingpolicies.OnCheckpointRollingPolicy;
+import org.apache.flink.table.api.Table;
+import org.apache.flink.table.api.bridge.java.StreamTableEnvironment;
 import org.apache.flink.util.InstantiationUtil;
 import org.apache.flink.util.Preconditions;
 
+import org.apache.flink.shaded.jackson2.com.fasterxml.jackson.core.JsonParser;
 import org.apache.flink.shaded.jackson2.com.fasterxml.jackson.databind.ObjectMapper;
 
 import java.io.BufferedReader;
 import java.io.BufferedWriter;
 import java.io.File;
-import java.io.FileReader;
-import java.io.FileWriter;
 import java.io.IOException;
+import java.io.InputStreamReader;
+import java.io.OutputStreamWriter;
 import java.lang.reflect.InvocationTargetException;
 import java.lang.reflect.Method;
-import java.nio.file.Path;
-import java.nio.file.Paths;
 import java.util.ArrayList;
 import java.util.Comparator;
 import java.util.HashMap;
@@ -58,22 +63,23 @@ import java.util.Set;
 
 /** Utility methods for reading and writing stages. */
 public class ReadWriteUtils {
-    public static final ObjectMapper OBJECT_MAPPER = new ObjectMapper();
+    public static final ObjectMapper OBJECT_MAPPER =
+            new ObjectMapper().enable(JsonParser.Feature.ALLOW_COMMENTS);
 
     // A helper method that calls encodes the given parameter value to a json string. We can not
     // call param.jsonEncode(value) directly because Param::jsonEncode(...) needs the actual type
     // of the value.
-    private static <T> String jsonEncodeHelper(Param<T> param, Object value) throws IOException {
+    private static <T> Object jsonEncodeHelper(Param<T> param, Object value) throws IOException {
         return param.jsonEncode((T) value);
     }
 
-    // Converts Map<Param<?>, Object> to Map<String, String> which maps the parameter name to the
-    // string-encoded parameter value.
-    private static Map<String, String> jsonEncode(Map<Param<?>, Object> paramMap)
+    // Converts Map<Param<?>, Object> to Map<String, Object> which maps the parameter name to the
+    // json-supported parameter value.
+    private static Map<String, Object> jsonEncode(Map<Param<?>, Object> paramMap)
             throws IOException {
-        Map<String, String> result = new HashMap<>(paramMap.size());
+        Map<String, Object> result = new HashMap<>(paramMap.size());
         for (Map.Entry<Param<?>, Object> entry : paramMap.entrySet()) {
-            String json = jsonEncodeHelper(entry.getKey(), entry.getValue());
+            Object json = jsonEncodeHelper(entry.getKey(), entry.getValue());
             result.put(entry.getKey().name, json);
         }
         return result;
@@ -91,9 +97,6 @@ public class ReadWriteUtils {
      */
     public static void saveMetadata(Stage<?> stage, String path, Map<String, ?> extraMetadata)
             throws IOException {
-        // Creates parent directories if not already created.
-        new File(path).mkdirs();
-
         Map<String, Object> metadata = new HashMap<>(extraMetadata);
         metadata.put("className", stage.getClass().getName());
         metadata.put("timestamp", System.currentTimeMillis());
@@ -101,12 +104,27 @@ public class ReadWriteUtils {
         // TODO: add version in the metadata.
         String metadataStr = OBJECT_MAPPER.writeValueAsString(metadata);
 
-        File metadataFile = new File(path, "metadata");
-        if (!metadataFile.createNewFile()) {
-            throw new IOException("File " + metadataFile.toString() + " already exists.");
+        saveToFile(new Path(path, "metadata").toUri().toString(), metadataStr, false);
+    }
+
+    /** Saves a given string to the specified file. */
+    public static void saveToFile(String pathStr, String content, boolean isOverwrite)
+            throws IOException {
+        Path path = new Path(pathStr);
+
+        // Creates parent directories if not already created.
+        FileSystem fs = mkdirs(path.getParent());
+
+        FileSystem.WriteMode writeMode = FileSystem.WriteMode.OVERWRITE;
+        if (!isOverwrite) {
+            writeMode = FileSystem.WriteMode.NO_OVERWRITE;
+            if (fs.exists(path)) {
+                throw new IOException("File " + path + " already exists.");
+            }
         }
-        try (BufferedWriter writer = new BufferedWriter(new FileWriter(metadataFile))) {
-            writer.write(metadataStr);
+        try (BufferedWriter writer =
+                new BufferedWriter(new OutputStreamWriter(fs.create(path, writeMode)))) {
+            writer.write(content);
         }
     }
 
@@ -125,20 +143,7 @@ public class ReadWriteUtils {
 
     /** Returns a subdirectory of the given path for saving/loading model data. */
     private static String getDataPath(String path) {
-        return Paths.get(path, "data").toString();
-    }
-
-    /** Returns all data files under the given path as a list of paths. */
-    private static org.apache.flink.core.fs.Path[] getDataPaths(String path) {
-        String dataPath = getDataPath(path);
-        File[] files = new File(dataPath).listFiles();
-
-        org.apache.flink.core.fs.Path[] paths = new org.apache.flink.core.fs.Path[files.length];
-        for (int i = 0; i < paths.length; i++) {
-            paths[i] = org.apache.flink.core.fs.Path.fromLocalFile(files[i]);
-        }
-
-        return paths;
+        return new Path(path, "data").toString();
     }
 
     /**
@@ -153,9 +158,11 @@ public class ReadWriteUtils {
      */
     public static Map<String, ?> loadMetadata(String path, String expectedClassName)
             throws IOException {
-        Path metadataPath = Paths.get(path, "metadata");
+        Path metadataPath = new Path(path, "metadata");
+        FileSystem fs = metadataPath.getFileSystem();
+
         StringBuilder buffer = new StringBuilder();
-        try (BufferedReader br = new BufferedReader(new FileReader(metadataPath.toString()))) {
+        try (BufferedReader br = new BufferedReader(new InputStreamReader(fs.open(metadataPath)))) {
             String line;
             while ((line = br.readLine()) != null) {
                 if (!line.startsWith("#")) {
@@ -184,9 +191,10 @@ public class ReadWriteUtils {
     // with zero or more `0` to have the same length as numStages. The resulting string can be
     // used as the directory to save a stage of the Pipeline or PipelineModel.
     private static String getPathForPipelineStage(int stageIdx, int numStages, String parentPath) {
-        String format = String.format("%%0%dd", String.valueOf(numStages).length());
+        String format =
+                String.format("stages%s%%0%dd", File.separator, String.valueOf(numStages).length());
         String fileName = String.format(format, stageIdx);
-        return Paths.get(parentPath, "stages", fileName).toString();
+        return new Path(parentPath, fileName).toString();
     }
 
     /**
@@ -199,7 +207,7 @@ public class ReadWriteUtils {
     public static void savePipeline(Stage<?> pipeline, List<Stage<?>> stages, String path)
             throws IOException {
         // Creates parent directories if not already created.
-        new File(path).mkdirs();
+        mkdirs(new Path(path));
 
         Map<String, Object> extraMetadata = new HashMap<>();
         extraMetadata.put("numStages", stages.size());
@@ -218,25 +226,30 @@ public class ReadWriteUtils {
      * <p>The method throws RuntimeException if the expectedClassName is not empty AND it does not
      * match the className of the previously saved Pipeline or PipelineModel.
      *
-     * @param env A StreamExecutionEnvironment instance.
+     * @param tEnv A StreamTableEnvironment instance.
      * @param path The parent directory to load the pipeline metadata and its stages.
      * @param expectedClassName The expected class name of the pipeline.
      * @return A list of stages.
      */
     public static List<Stage<?>> loadPipeline(
-            StreamExecutionEnvironment env, String path, String expectedClassName)
-            throws IOException {
+            StreamTableEnvironment tEnv, String path, String expectedClassName) throws IOException {
         Map<String, ?> metadata = loadMetadata(path, expectedClassName);
         int numStages = (Integer) metadata.get("numStages");
         List<Stage<?>> stages = new ArrayList<>(numStages);
 
         for (int i = 0; i < numStages; i++) {
             String stagePath = getPathForPipelineStage(i, numStages, path);
-            stages.add(loadStage(env, stagePath));
+            stages.add(loadStage(tEnv, stagePath));
         }
         return stages;
     }
 
+    private static FileSystem mkdirs(Path path) throws IOException {
+        FileSystem fs = path.getFileSystem();
+        fs.mkdirs(path);
+        return fs;
+    }
+
     /**
      * Saves a Graph or GraphModel with the given GraphData to the given path.
      *
@@ -247,7 +260,7 @@ public class ReadWriteUtils {
     public static void saveGraph(Stage<?> graph, GraphData graphData, String path)
             throws IOException {
         // Creates parent directories if not already created.
-        new File(path).mkdirs();
+        mkdirs(new Path(path));
 
         Map<String, Object> extraMetadata = new HashMap<>();
         extraMetadata.put("graphData", graphData.toMap());
@@ -270,14 +283,14 @@ public class ReadWriteUtils {
      * <p>The method throws RuntimeException if the expectedClassName is not empty AND it does not
      * match the className of the previously saved Pipeline or PipelineModel.
      *
-     * @param env A StreamExecutionEnvironment instance.
+     * @param tEnv A StreamTableEnvironment instance.
      * @param path The parent directory to load the pipeline metadata and its stages.
      * @param expectedClassName The expected class name of the pipeline.
      * @return A Graph or GraphModel instance.
      */
+    @SuppressWarnings("unchecked")
     public static Stage<?> loadGraph(
-            StreamExecutionEnvironment env, String path, String expectedClassName)
-            throws IOException {
+            StreamTableEnvironment tEnv, String path, String expectedClassName) throws IOException {
         Map<String, ?> metadata = loadMetadata(path, expectedClassName);
         GraphData graphData = GraphData.fromMap((Map<String, Object>) metadata.get("graphData"));
 
@@ -289,7 +302,7 @@ public class ReadWriteUtils {
 
         for (GraphNode node : graphData.nodes) {
             String stagePath = getPathForPipelineStage(node.nodeId, maxNodeId + 1, path);
-            node.stage = loadStage(env, stagePath);
+            node.stage = loadStage(tEnv, stagePath);
         }
 
         if (expectedClassName.equals(GraphModel.class.getName())) {
@@ -310,20 +323,23 @@ public class ReadWriteUtils {
                 graphData.outputModelDataIds);
     }
 
-    // A helper method that sets stage's parameter value. We can not call stage.set(param, value)
-    // directly because stage::set(...) needs the actual type of the value.
-    public static <T> void setParam(Stage<?> stage, Param<T> param, Object value) {
-        stage.set(param, (T) value);
+    // A helper method that sets WithParams object's parameter value. We can not call
+    // WithParams.set(param, value)
+    // directly because WithParams::set(...) needs the actual type of the value.
+    @SuppressWarnings("unchecked")
+    public static <T> void setParam(WithParams<?> instance, Param<T> param, Object value) {
+        instance.set(param, (T) value);
     }
 
-    // A helper method that updates stage's param map using values from the paramOverrides. This
-    // method only
-    // updates values for parameters already defined in the stage's param map.
-    public static void updateExistingParams(Stage<?> stage, Map<Param<?>, Object> paramOverrides) {
-        Set<Param<?>> existingParams = stage.getParamMap().keySet();
+    // A helper method that updates WithParams instance's param map using values from the
+    // paramOverrides. This method only updates values for parameters already defined in the
+    // instance's param map.
+    public static void updateExistingParams(
+            WithParams<?> instance, Map<Param<?>, Object> paramOverrides) {
+        Set<Param<?>> existingParams = instance.getParamMap().keySet();
         for (Map.Entry<Param<?>, Object> entry : paramOverrides.entrySet()) {
             if (existingParams.contains(entry.getKey())) {
-                setParam(stage, entry.getKey(), entry.getValue());
+                setParam(instance, entry.getKey(), entry.getValue());
             }
         }
     }
@@ -342,31 +358,43 @@ public class ReadWriteUtils {
      * @param <T> The class type of the Stage subclass.
      * @return An instance of class type T.
      */
-    @SuppressWarnings("unchecked")
     public static <T extends Stage<T>> T loadStageParam(String path) throws IOException {
-        Map<String, ?> metadata = loadMetadata(path, "");
-        String className = (String) metadata.get("className");
-        Map<String, String> paramMap = (Map<String, String>) metadata.get("paramMap");
-
         try {
-            Class<T> clazz = (Class<T>) Class.forName(className);
-            T instance = InstantiationUtil.instantiate(clazz);
-
-            Map<String, Param<?>> nameToParam = new HashMap<>();
-            for (Param<?> param : ParamUtils.getPublicFinalParamFields(instance)) {
-                nameToParam.put(param.name, param);
-            }
-
-            for (Map.Entry<String, String> entry : paramMap.entrySet()) {
-                Param<?> param = nameToParam.get(entry.getKey());
-                setParam(instance, param, param.jsonDecode(entry.getValue()));
-            }
-            return instance;
+            return instantiateWithParams(loadMetadata(path, ""));
         } catch (ClassNotFoundException e) {
             throw new RuntimeException("Failed to load stage.", e);
         }
     }
 
+    /**
+     * Instantiates a WithParams subclass from the provided json map.
+     *
+     * @param jsonMap a map containing className and paramMap.
+     * @return the instantiated WithParams subclass instance.
+     */
+    @SuppressWarnings("unchecked")
+    public static <T extends WithParams<T>> T instantiateWithParams(Map<String, ?> jsonMap)
+            throws ClassNotFoundException, IOException {
+        String className = (String) jsonMap.get("className");
+        Class<T> clazz = (Class<T>) Class.forName(className);
+        T instance = InstantiationUtil.instantiate(clazz);
+
+        Map<String, Param<?>> nameToParam = new HashMap<>();
+        for (Param<?> param : ParamUtils.getPublicFinalParamFields(instance)) {
+            nameToParam.put(param.name, param);
+        }
+
+        if (jsonMap.containsKey("paramMap")) {
+            Map<String, Object> paramMap = (Map<String, Object>) jsonMap.get("paramMap");
+            for (Map.Entry<String, Object> entry : paramMap.entrySet()) {
+                Param<?> param = nameToParam.get(entry.getKey());
+                setParam(instance, param, param.jsonDecode(entry.getValue()));
+            }
+        }
+
+        return instance;
+    }
+
     /**
      * Loads the stage from the given path by invoking the static load() method of the stage. The
      * stage class name is read from the metadata file under the given path. The load() method is
@@ -375,20 +403,19 @@ public class ReadWriteUtils {
      *
      * <p>Required: the stage class must have a static load() method.
      *
-     * @param env A StreamExecutionEnvironment instance.
+     * @param tEnv A StreamTableEnvironment instance.
      * @param path The parent directory of the stage metadata file.
      * @return An instance of Stage.
      */
-    public static Stage<?> loadStage(StreamExecutionEnvironment env, String path)
-            throws IOException {
+    public static Stage<?> loadStage(StreamTableEnvironment tEnv, String path) throws IOException {
         Map<String, ?> metadata = loadMetadata(path, "");
         String className = (String) metadata.get("className");
 
         try {
             Class<?> clazz = Class.forName(className);
-            Method method = clazz.getMethod("load", StreamExecutionEnvironment.class, String.class);
+            Method method = clazz.getMethod("load", StreamTableEnvironment.class, String.class);
             method.setAccessible(true);
-            return (Stage<?>) method.invoke(null, env, path);
+            return (Stage<?>) method.invoke(null, tEnv, path);
         } catch (NoSuchMethodException e) {
             String methodName = String.format("%s::load(String)", className);
             throw new RuntimeException(
@@ -423,16 +450,19 @@ public class ReadWriteUtils {
     /**
      * Loads the model data from the given path using the model decoder.
      *
-     * @param env A StreamExecutionEnvironment instance.
+     * @param tEnv A StreamTableEnvironment instance.
      * @param path The parent directory of the model data file.
      * @param modelDecoder The decoder used to decode the model data.
      * @param <T> The class type of the model data.
      * @return The loaded model data.
      */
-    public static <T> DataStream<T> loadModelData(
-            StreamExecutionEnvironment env, String path, SimpleStreamFormat<T> modelDecoder) {
+    public static <T> Table loadModelData(
+            StreamTableEnvironment tEnv, String path, SimpleStreamFormat<T> modelDecoder) {
+        StreamExecutionEnvironment env = TableUtils.getExecutionEnvironment(tEnv);
         Source<T, ?, ?> source =
-                FileSource.forRecordStreamFormat(modelDecoder, getDataPaths(path)).build();
-        return env.fromSource(source, WatermarkStrategy.noWatermarks(), "modelData");
+                FileSource.forRecordStreamFormat(modelDecoder, new Path(getDataPath(path))).build();
+        DataStream<T> modelDataStream =
+                env.fromSource(source, WatermarkStrategy.noWatermarks(), "modelData");
+        return tEnv.fromDataStream(modelDataStream);
     }
 }
