@@ -19,6 +19,7 @@
 package org.apache.flink.ml.api;
 
 import org.apache.flink.ml.param.BooleanParam;
+import org.apache.flink.ml.param.DoubleArrayArrayParam;
 import org.apache.flink.ml.param.DoubleArrayParam;
 import org.apache.flink.ml.param.DoubleParam;
 import org.apache.flink.ml.param.FloatArrayParam;
@@ -30,12 +31,14 @@ import org.apache.flink.ml.param.LongParam;
 import org.apache.flink.ml.param.Param;
 import org.apache.flink.ml.param.ParamValidator;
 import org.apache.flink.ml.param.ParamValidators;
+import org.apache.flink.ml.param.StringArrayArrayParam;
 import org.apache.flink.ml.param.StringArrayParam;
 import org.apache.flink.ml.param.StringParam;
 import org.apache.flink.ml.param.WithParams;
 import org.apache.flink.ml.util.ParamUtils;
 import org.apache.flink.ml.util.ReadWriteUtils;
 import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
+import org.apache.flink.table.api.bridge.java.StreamTableEnvironment;
 
 import org.junit.Assert;
 import org.junit.Test;
@@ -56,9 +59,18 @@ public class StageTest {
         Param<Integer> INT_PARAM =
                 new IntParam("intParam", "Description", 1, ParamValidators.lt(100));
 
+        // This param might be regarded as integer type given a small value.
         Param<Long> LONG_PARAM =
                 new LongParam("longParam", "Description", 2L, ParamValidators.lt(100));
 
+        // This param must be regarded as long type given a large enough value.
+        Param<Long> LONG_PARAM2 =
+                new LongParam(
+                        "longParam2",
+                        "Description",
+                        Integer.MAX_VALUE + 1L,
+                        ParamValidators.lt(Integer.MAX_VALUE + 100L));
+
         Param<Float> FLOAT_PARAM =
                 new FloatParam("floatParam", "Description", 3.0f, ParamValidators.lt(100));
 
@@ -89,6 +101,18 @@ public class StageTest {
 
         Param<String[]> STRING_ARRAY_PARAM =
                 new StringArrayParam("stringArrayParam", "Description", new String[] {"14", "15"});
+
+        Param<String[][]> STRING_ARRAY_ARRAY_PARAM =
+                new StringArrayArrayParam(
+                        "stringArrayArrayParam",
+                        "Description",
+                        new String[][] {new String[] {"14", "15"}});
+
+        Param<Double[][]> DOUBLE_ARRAY_ARRAY_PARAM =
+                new DoubleArrayArrayParam(
+                        "doubleArrayArrayParam",
+                        "Description",
+                        new Double[][] {new Double[] {14.0, 15.0}, new Double[] {16.0, 17.0}});
     }
 
     /**
@@ -121,7 +145,7 @@ public class StageTest {
             ReadWriteUtils.saveMetadata(this, path);
         }
 
-        public static MyStage load(StreamExecutionEnvironment env, String path) throws IOException {
+        public static MyStage load(StreamTableEnvironment tEnv, String path) throws IOException {
             return ReadWriteUtils.loadStageParam(path);
         }
     }
@@ -167,7 +191,7 @@ public class StageTest {
     // Saves and loads the given stage. And verifies that the loaded stage has same parameter values
     // as the original stage.
     private static Stage<?> validateStageSaveLoad(
-            StreamExecutionEnvironment env, Stage<?> stage, Map<String, Object> paramOverrides)
+            StreamTableEnvironment tEnv, Stage<?> stage, Map<String, Object> paramOverrides)
             throws IOException {
         for (Map.Entry<String, Object> entry : paramOverrides.entrySet()) {
             Param<?> param = stage.getParam(entry.getKey());
@@ -183,7 +207,7 @@ public class StageTest {
             // This is expected.
         }
 
-        Stage<?> loadedStage = ReadWriteUtils.loadStage(env, path);
+        Stage<?> loadedStage = ReadWriteUtils.loadStage(tEnv, path);
         for (Map.Entry<String, Object> entry : paramOverrides.entrySet()) {
             Param<?> param = loadedStage.getParam(entry.getKey());
             Assert.assertEquals(entry.getValue(), loadedStage.get(param));
@@ -258,6 +282,7 @@ public class StageTest {
         MyStage stage = new MyStage();
         assertInvalidValue(stage, MyParams.INT_PARAM, 100);
         assertInvalidValue(stage, MyParams.LONG_PARAM, 100L);
+        assertInvalidValue(stage, MyParams.LONG_PARAM2, Integer.MAX_VALUE + 100L);
         assertInvalidValue(stage, MyParams.FLOAT_PARAM, 100.0f);
         assertInvalidValue(stage, MyParams.DOUBLE_PARAM, 100.0);
         assertInvalidClass(stage, MyParams.INT_PARAM, "100");
@@ -280,11 +305,14 @@ public class StageTest {
         stage.set(MyParams.LONG_PARAM, 50L);
         Assert.assertEquals(50L, (long) stage.get(MyParams.LONG_PARAM));
 
+        stage.set(MyParams.LONG_PARAM2, Integer.MAX_VALUE + 50L);
+        Assert.assertEquals(Integer.MAX_VALUE + 50L, (long) stage.get(MyParams.LONG_PARAM2));
+
         stage.set(MyParams.FLOAT_PARAM, 50f);
-        Assert.assertEquals(50f, (float) stage.get(MyParams.FLOAT_PARAM), 0.0001);
+        Assert.assertEquals(50f, stage.get(MyParams.FLOAT_PARAM), 0.0001);
 
         stage.set(MyParams.DOUBLE_PARAM, 50.0);
-        Assert.assertEquals(50, (double) stage.get(MyParams.DOUBLE_PARAM), 0.0001);
+        Assert.assertEquals(50, stage.get(MyParams.DOUBLE_PARAM), 0.0001);
 
         stage.set(MyParams.STRING_PARAM, "50");
         Assert.assertEquals("50", stage.get(MyParams.STRING_PARAM));
@@ -303,33 +331,107 @@ public class StageTest {
 
         stage.set(MyParams.STRING_ARRAY_PARAM, new String[] {"50", "51"});
         Assert.assertArrayEquals(new String[] {"50", "51"}, stage.get(MyParams.STRING_ARRAY_PARAM));
+
+        stage.set(
+                MyParams.DOUBLE_ARRAY_ARRAY_PARAM,
+                new Double[][] {new Double[] {50.0, 51.0}, new Double[] {52.0, 53.0}});
+        Assert.assertEquals(2, stage.get(MyParams.DOUBLE_ARRAY_ARRAY_PARAM).length);
+        Assert.assertArrayEquals(
+                new Double[] {50.0, 51.0}, stage.get(MyParams.DOUBLE_ARRAY_ARRAY_PARAM)[0]);
+        Assert.assertArrayEquals(
+                new Double[] {52.0, 53.0}, stage.get(MyParams.DOUBLE_ARRAY_ARRAY_PARAM)[1]);
+
+        stage.set(
+                MyParams.STRING_ARRAY_ARRAY_PARAM,
+                new String[][] {
+                    new String[] {"50", "51"},
+                    new String[] {"52", "53"}
+                });
+        Assert.assertEquals(2, stage.get(MyParams.STRING_ARRAY_ARRAY_PARAM).length);
+        Assert.assertArrayEquals(
+                new String[] {"50", "51"}, stage.get(MyParams.STRING_ARRAY_ARRAY_PARAM)[0]);
+        Assert.assertArrayEquals(
+                new String[] {"52", "53"}, stage.get(MyParams.STRING_ARRAY_ARRAY_PARAM)[1]);
     }
 
     @Test
     public void testStageSaveLoad() throws IOException {
         StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
+        StreamTableEnvironment tEnv = StreamTableEnvironment.create(env);
         MyStage stage = new MyStage();
+
         stage.set(stage.paramWithNullDefault, 1);
-        Stage<?> loadedStage = validateStageSaveLoad(env, stage, Collections.emptyMap());
-        Assert.assertEquals(1, (int) loadedStage.get(MyParams.INT_PARAM));
+        stage.set(MyParams.BOOLEAN_PARAM, true);
+        stage.set(MyParams.INT_PARAM, 50);
+        stage.set(MyParams.LONG_PARAM, 50L);
+        stage.set(MyParams.LONG_PARAM2, Integer.MAX_VALUE + 50L);
+        stage.set(MyParams.FLOAT_PARAM, 50f);
+        stage.set(MyParams.DOUBLE_PARAM, 50.0);
+        stage.set(MyParams.STRING_PARAM, "50");
+        stage.set(MyParams.INT_ARRAY_PARAM, new Integer[] {50, 51});
+        stage.set(MyParams.LONG_ARRAY_PARAM, new Long[] {50L, 51L});
+        stage.set(MyParams.FLOAT_ARRAY_PARAM, new Float[] {50.0f, 51.0f});
+        stage.set(MyParams.DOUBLE_ARRAY_PARAM, new Double[] {50.0, 51.0});
+        stage.set(MyParams.STRING_ARRAY_PARAM, new String[] {"50", "51"});
+        stage.set(
+                MyParams.DOUBLE_ARRAY_ARRAY_PARAM,
+                new Double[][] {new Double[] {50.0, 51.0}, new Double[] {52.0, 53.0}});
+        stage.set(
+                MyParams.STRING_ARRAY_ARRAY_PARAM,
+                new String[][] {
+                    new String[] {"50", "51"},
+                    new String[] {"52", "53"}
+                });
+
+        Stage<?> loadedStage = validateStageSaveLoad(tEnv, stage, Collections.emptyMap());
+
+        Assert.assertEquals(1, (int) loadedStage.get(stage.paramWithNullDefault));
+        Assert.assertEquals(true, loadedStage.get(MyParams.BOOLEAN_PARAM));
+        Assert.assertEquals(50, (int) loadedStage.get(MyParams.INT_PARAM));
+        Assert.assertEquals(50L, (long) loadedStage.get(MyParams.LONG_PARAM));
+        Assert.assertEquals(Integer.MAX_VALUE + 50L, (long) loadedStage.get(MyParams.LONG_PARAM2));
+        Assert.assertEquals(50f, loadedStage.get(MyParams.FLOAT_PARAM), 0.0001);
+        Assert.assertEquals(50, loadedStage.get(MyParams.DOUBLE_PARAM), 0.0001);
+        Assert.assertEquals("50", loadedStage.get(MyParams.STRING_PARAM));
+        Assert.assertArrayEquals(new Integer[] {50, 51}, loadedStage.get(MyParams.INT_ARRAY_PARAM));
+        Assert.assertArrayEquals(new Long[] {50L, 51L}, loadedStage.get(MyParams.LONG_ARRAY_PARAM));
+        Assert.assertArrayEquals(
+                new Float[] {50.0f, 51.0f}, loadedStage.get(MyParams.FLOAT_ARRAY_PARAM));
+        Assert.assertArrayEquals(
+                new Double[] {50.0, 51.0}, loadedStage.get(MyParams.DOUBLE_ARRAY_PARAM));
+        Assert.assertArrayEquals(
+                new String[] {"50", "51"}, loadedStage.get(MyParams.STRING_ARRAY_PARAM));
+        Assert.assertEquals(2, stage.get(MyParams.DOUBLE_ARRAY_ARRAY_PARAM).length);
+        Assert.assertArrayEquals(
+                new Double[] {50.0, 51.0}, stage.get(MyParams.DOUBLE_ARRAY_ARRAY_PARAM)[0]);
+        Assert.assertArrayEquals(
+                new Double[] {52.0, 53.0}, stage.get(MyParams.DOUBLE_ARRAY_ARRAY_PARAM)[1]);
+        Assert.assertEquals(2, stage.get(MyParams.STRING_ARRAY_ARRAY_PARAM).length);
+        Assert.assertArrayEquals(
+                new String[] {"50", "51"}, stage.get(MyParams.STRING_ARRAY_ARRAY_PARAM)[0]);
+        Assert.assertArrayEquals(
+                new String[] {"52", "53"}, stage.get(MyParams.STRING_ARRAY_ARRAY_PARAM)[1]);
     }
 
     @Test
     public void testStageSaveLoadWithParamOverrides() throws IOException {
         StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
+        StreamTableEnvironment tEnv = StreamTableEnvironment.create(env);
         MyStage stage = new MyStage();
         stage.set(stage.paramWithNullDefault, 1);
         Stage<?> loadedStage =
-                validateStageSaveLoad(env, stage, Collections.singletonMap("intParam", 10));
-        Assert.assertEquals(10, (int) loadedStage.get(MyParams.INT_PARAM));
+                validateStageSaveLoad(
+                        tEnv, stage, Collections.singletonMap("paramWithNullDefault", 10));
+        Assert.assertEquals(10, (int) loadedStage.get(stage.paramWithNullDefault));
     }
 
     @Test
     public void testStageLoadWithoutLoadMethod() throws IOException {
         StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
+        StreamTableEnvironment tEnv = StreamTableEnvironment.create(env);
         MyStageWithoutLoad stage = new MyStageWithoutLoad();
         try {
-            validateStageSaveLoad(env, stage, Collections.emptyMap());
+            validateStageSaveLoad(tEnv, stage, Collections.emptyMap());
             Assert.fail("Expected RuntimeException");
         } catch (RuntimeException e) {
             Assert.assertTrue(e.getMessage().contains("not implemented"));
@@ -391,5 +493,11 @@ public class StageTest {
         Assert.assertTrue(nonEmptyArray.validate(new String[] {"1"}));
         Assert.assertFalse(nonEmptyArray.validate(null));
         Assert.assertFalse(nonEmptyArray.validate(new String[0]));
+
+        ParamValidator<String[]> isSubArray = ParamValidators.isSubSet("a", "b", "c");
+        Assert.assertFalse(isSubArray.validate(null));
+        Assert.assertFalse(isSubArray.validate(new String[] {"c", "v"}));
+        Assert.assertTrue(isSubArray.validate(new String[] {"a", "b"}));
+        Assert.assertFalse(isSubArray.validate(new String[] {"e", "v"}));
     }
 }
